package org.eclipse.weaver.extension.byADL.transformation;

import java.net.URL;
import java.util.ArrayList;
import java.util.Collections;
import java.util.Date;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Map;

import org.eclipse.core.resources.IFolder;
import org.eclipse.core.resources.IProject;
import org.eclipse.core.resources.IResource;
import org.eclipse.core.runtime.FileLocator;
import org.eclipse.core.runtime.IStatus;
import org.eclipse.core.runtime.Platform;
import org.eclipse.emf.common.util.URI;
import org.eclipse.emf.ecore.EAttribute;
import org.eclipse.emf.ecore.EClassifier;
import org.eclipse.emf.ecore.EPackage;
import org.eclipse.emf.ecore.impl.DynamicEObjectImpl;
import org.eclipse.emf.ecore.resource.Resource;
import org.eclipse.emf.ecore.resource.ResourceSet;
import org.eclipse.emf.ecore.resource.impl.ResourceSetImpl;
import org.eclipse.emf.ecore.xmi.impl.XMLResourceFactoryImpl;
import org.eclipse.gmt.weaver.amw4atl.ASMAMWModel;
import org.eclipse.gmt.weaver.transformation.util.TransformationConstant;
import org.eclipse.m2m.atl.drivers.emf4atl.ASMEMFModel;
import org.eclipse.m2m.atl.drivers.emf4atl.ASMEMFModelElement;
import org.eclipse.m2m.atl.engine.AtlEMFModelHandler;
import org.eclipse.m2m.atl.engine.AtlLauncher;
import org.eclipse.m2m.atl.engine.AtlModelHandler;
import org.eclipse.m2m.atl.engine.vm.nativelib.ASMModel;
import org.eclipse.weaver.core.WeaverModelManager;
import org.eclipse.weaver.core.WeaverXMLMetadata;
import org.eclipse.weaver.exception.WeaverException;

/**
 * This class manages the execution, management and retrieve of ATL transformations.
 * It is inspired by a class of the transformations plug-in of AMW.
 * 
 * @author - Ivano Malavolta (Department of Computer Science)<br>
 * 		   - Universit dell'Aquila (Italy)
 * @since 1.0
 */
public class TransformationsManager {

	/**
	 * this list contains the Resources of the woven models
	 */
	public static List<Resource> wovenModels = new ArrayList<Resource>();

	/**
	 * this object stores all the informations to execute and manage the 'composition' transformation
	 */
	private static TransformationConfig compositionTransformation = TransformationsManager.createCompositionTransformation();
	
	/**
	 * this object stores all the informations to execute and manage the 'composition' transformation
	 */
	private static TransformationConfig single2ComposedGenerator = TransformationsManager.createSingle2ComposedGenerator();
	
	/**
	 * this object stores all the informations to execute and manage the 'composition' transformation
	 */
	private static TransformationConfig composed2SingleGenerator = TransformationsManager.createComposed2SingleGenerator();

	/**
	 * executes a transformation.
	 * @param models the map of models
	 * @param wManager the model manager
	 * @param fileXML the xml file associated with the weaving model that is being edited
	 * @param tc the configuration of the transformation that will be executed
	 * @return
	 * @throws WeaverException 
	 */
	public static ASMModel executeCompositionTransfo(Map models, WeaverModelManager wManager,
			WeaverXMLMetadata fileXML, TransformationConfig tc) throws WeaverException {
		ASMModel ret = null;   
		try { 
			AtlModelHandler modelHandler = AtlModelHandler.getDefault("AMW");
			URL tPlatform = Platform.getBundle(tc.plugin).getEntry(tc.transformation);	
			// the URL of the transformation
			URL transfoURL = null;
			transfoURL = FileLocator.toFileURL(tPlatform);
			if (models.get("MOF") == null) {
				models.put("MOF", modelHandler.getMof());
			}
			if (models.get("ECORE") == null) {
				models.put("ECORE", modelHandler.getMof());
			}
			//now get the woven models
			Iterator wModels = TransformationsManager.wovenModels.iterator();
			// Set the variable that represents the left woven model
			Resource woven = (Resource) wModels.next();
			ASMEMFModel leftWovenEMF = null;
			leftWovenEMF = ASMAMWModel.loadASMAMWModel("LEFT", (ASMEMFModel)modelHandler.getMof(), woven, null);
			models.put("LEFT", leftWovenEMF); 
			// Set the variable that represents the right woven model
			woven = (Resource) wModels.next();
			ASMEMFModel rightWovenEMF = null;
			rightWovenEMF = ASMAMWModel.loadASMAMWModel("RIGHT", (ASMEMFModel)modelHandler.getMof(), woven, null);
			//serializeMetamodel((EPackage) woven.getContents().get(0));
			models.put("RIGHT", rightWovenEMF);
			// this instruction executes the ATL transformation
			System.out.println((new Date()));
			Map params = new HashMap();
			params.put("AllowInterModelReferences", Boolean.TRUE);
//			Map options = new HashMap();
//			options.put("continueAfterError", "true");
			AtlLauncher.getDefault().launch(transfoURL, Collections.EMPTY_MAP, models, params, Collections.EMPTY_LIST, Collections.EMPTY_MAP);
			System.out.println((new Date()));
			// this is the output model
			ASMEMFModel out = (ASMEMFModel) models.get("OUT");
			ret = out;
		} catch (Exception e) {
			e.printStackTrace();
			throw new WeaverException(IStatus.ERROR,"The input or output models of the transformation could not be configured");
		}		
		return ret;
	}
	
	/**
	 * executes a transformation.
	 * @param models the map of models
	 * @param wManager the model manager
	 * @param fileXML the xml file associated with the weaving model that is being edited
	 * @param tc the configuration of the transformation that will be executed
	 * @return
	 * @throws WeaverException 
	 */
	public static ASMModel executeGeneratorsTransfo(Map models, WeaverModelManager wManager,
			WeaverXMLMetadata fileXML, TransformationConfig tc) throws WeaverException {
		ASMModel ret = null;   
		try { 
			AtlModelHandler modelHandler = AtlModelHandler.getDefault("AMW");
			URL tPlatform = Platform.getBundle(tc.plugin).getEntry(tc.transformation);	
			// the URL of the transformation
			URL transfoURL = null;
			transfoURL = FileLocator.toFileURL(tPlatform);
			if (models.get("MOF") == null) {
				models.put("MOF", modelHandler.getMof());
			}
			if (models.get("ECORE") == null) {
				models.put("ECORE", modelHandler.getMof());
			}
			//now get the woven models
			Iterator wModels = TransformationsManager.wovenModels.iterator();
			// Set the variable that represents the left woven model
			Resource woven = (Resource) wModels.next();
			ASMEMFModel leftWovenEMF = null;
			leftWovenEMF = ASMAMWModel.loadASMAMWModel("LEFT", (ASMEMFModel)modelHandler.getMof(), woven, null);
			models.put("LEFT", leftWovenEMF); 
			// Set the variable that represents the right woven model
			woven = (Resource) wModels.next();
			ASMEMFModel rightWovenEMF = null;
			rightWovenEMF = ASMAMWModel.loadASMAMWModel("RIGHT", (ASMEMFModel)modelHandler.getMof(), woven, null);
			models.put("RIGHT", rightWovenEMF);
			// set the variable representing the composed metamodel
			ASMEMFModel composedEMF = null;
			IFolder metamodelsFolder;
			IProject proj;
			URL mmurl;
			String inModelUri = wManager.getWeavingModel().getURI().trimSegments(2).path().replaceAll("resource/", "") + "/" + (((DynamicEObjectImpl) wManager.getWeavingModel().getContents().get(0)).eGet(((DynamicEObjectImpl) wManager.getWeavingModel().getContents().get(0)).eClass().getEStructuralFeature("name"))).toString() + ".ecore";
			if(org.eclipse.core.resources.ResourcesPlugin.getWorkspace().getRoot().getProject(inModelUri.split("/")[1]).findMember("metamodels") != null) {
				metamodelsFolder = (IFolder) org.eclipse.core.resources.ResourcesPlugin.getWorkspace().getRoot().getProject(inModelUri.split("/")[1]).findMember("metamodels");
				mmurl = metamodelsFolder.findMember((((DynamicEObjectImpl) wManager.getWeavingModel().getContents().get(0)).eGet(((DynamicEObjectImpl) wManager.getWeavingModel().getContents().get(0)).eClass().getEStructuralFeature("name"))).toString() + ".ecore", false).getLocationURI().toURL();
			} else {
				proj = org.eclipse.core.resources.ResourcesPlugin.getWorkspace().getRoot().getProject(wManager.getWeavingModel().getURI().toString().split("/")[1]);
				mmurl = proj.findMember((String) (((DynamicEObjectImpl) wManager.getWeavingModel().getContents().get(0)).eGet(((DynamicEObjectImpl) wManager.getWeavingModel().getContents().get(0)).eClass().getEStructuralFeature("name"))), false).getLocationURI().toURL();
			}
			composedEMF = (ASMEMFModel) AtlModelHandler.getDefault("EMF").loadModel("IN", (ASMEMFModel)modelHandler.getMof(), mmurl.openStream());
			models.put("IN", composedEMF);
			System.out.println((new Date()));
			Map params = new HashMap();
			params.put("AllowInterModelReferences", Boolean.TRUE);
			URL configPlatform = Platform.getBundle(tc.plugin).getEntry("transformations/Config.asm");	
			// the URL of the transformation
			URL configTransfoURL = null;
			configTransfoURL = FileLocator.toFileURL(configPlatform);
			Map libraries = new HashMap();
			libraries.put("Config", configTransfoURL);
//			Map options = new HashMap();
//			options.put("continueAfterError", "true");
			AtlLauncher.getDefault().launch(transfoURL, libraries, models, params, Collections.EMPTY_LIST, Collections.EMPTY_MAP);
			System.out.println((new Date()));
			// this is the output model
			ASMEMFModel out = (ASMEMFModel) models.get("OUT");
			if(out.getMetamodel().equals(AtlModelHandler.getDefault("AMW").getMof())) {
				fixAttributesTypes(out, leftWovenEMF, rightWovenEMF);
			}
			ret = out;
		} catch (Exception e) {
			e.printStackTrace();
			throw new WeaverException(IStatus.ERROR,"The input or output models of the transformation could not be configured");
		}		
		return ret;
	}
	
	//TODO remove this method and find out why primitive types are not set by a programmatically launched transformation
	private static void fixAttributesTypes(ASMEMFModel out, ASMEMFModel left, ASMEMFModel right) {
		EAttribute current;
		Map leftAttributes = new HashMap();
		Map rightAttributes = new HashMap();
		for(int i=0; i < left.getElementsByType("EAttribute").size(); i++) {
			current = (EAttribute)((ASMEMFModelElement) left.getElementsByType("EAttribute").toArray()[i]).getObject();
			leftAttributes.put(current.getName(), current.getEType());
		}
		for(int i=0; i < right.getElementsByType("EAttribute").size(); i++) {
			current = (EAttribute)((ASMEMFModelElement) right.getElementsByType("EAttribute").toArray()[i]).getObject();
			rightAttributes.put(current.getName(), current.getEType());
		}
		for(int i=0; i < out.getElementsByType("EAttribute").size(); i++) {
			current = (EAttribute)((ASMEMFModelElement) out.getElementsByType("EAttribute").toArray()[i]).getObject();
			if(current.getEType() == null) {
				if(leftAttributes.containsKey(current.getName())) {
					current.setEType((EClassifier) leftAttributes.get(current.getName()));
				} else {
					current.setEType((EClassifier) rightAttributes.get(current.getName()));
				}
			}
		}
	}
	
	/*
	 * @param tc
	 * @param in
	 * @param wManager
	 * @return
	 */
	public static ASMModel saveNewModel (TransformationConfig tc, ASMModel in, WeaverModelManager wManager, WeaverXMLMetadata currentXML) {
		ASMModel output = null;
		try {	    		
			Resource wModel = wManager.getWeavingModel();
			AtlModelHandler modelHandler = (AtlEMFModelHandler) AtlEMFModelHandler.getDefault(AtlEMFModelHandler.AMH_EMF);
			String inModelUri;
			String fileName = (String) (((DynamicEObjectImpl) wModel.getContents().get(0)).eGet(((DynamicEObjectImpl) wModel.getContents().get(0)).eClass().getEStructuralFeature("name")));
			inModelUri = wModel.getURI().trimSegments(2).path().replaceAll("resource/", "") + "/" + fileName + tc.suffix + ".ecore";
			IProject proj = org.eclipse.core.resources.ResourcesPlugin.getWorkspace().getRoot().getProject(inModelUri.split("/")[1]);
			// if the current project contains a folder named "transformations", then save the model transformation there
			if(proj.findMember("metamodels") != null && proj.findMember("metamodels").getType() == IResource.FOLDER) {
				inModelUri = "/" + proj.getName() + "/metamodels/" + inModelUri.split("/")[inModelUri.split("/").length -1];
			}
			modelHandler.saveModel(in, inModelUri);
			output = in;
		} catch (Exception e) {
			e.printStackTrace();
		}		
		return output;
	}
	
	/*
	 * @param tc
	 * @param in
	 * @param wManager
	 * @return
	 */
	public static ASMModel saveNewMigrator (TransformationConfig tc, ASMModel in, WeaverModelManager wManager, WeaverXMLMetadata currentXML) {
		ASMModel output = null;
		try {	    		
			Resource wModel = wManager.getWeavingModel();
			AtlModelHandler modelHandler = (AtlEMFModelHandler) AtlEMFModelHandler.getDefault(AtlEMFModelHandler.AMH_EMF);
			String inModelUri;
			String fileName = (String) (((DynamicEObjectImpl) wModel.getContents().get(0)).eGet(((DynamicEObjectImpl) wModel.getContents().get(0)).eClass().getEStructuralFeature("name")));
			inModelUri = wModel.getURI().trimSegments(2).path().replaceAll("resource/", "") + "/" + fileName + tc.suffix + ".ecore";
			IProject proj = org.eclipse.core.resources.ResourcesPlugin.getWorkspace().getRoot().getProject(inModelUri.split("/")[1]);
			// if the current project contains a folder named "transformations", then save the model transformation there
			if(proj.findMember("generatedMigrators") != null && proj.findMember("generatedMigrators").getType() == IResource.FOLDER) {
				inModelUri = "/" + proj.getName() + "/generatedMigrators/" + inModelUri.split("/")[inModelUri.split("/").length -1];
			} else {			
				if(proj.findMember("transformations") != null && proj.findMember("transformations").getType() == IResource.FOLDER) {
					inModelUri = "/" + proj.getName() + "/transformations/" + inModelUri.split("/")[inModelUri.split("/").length -1];
				}
			}
			modelHandler.saveModel(in, inModelUri);
			output = in;
		} catch (Exception e) {
			e.printStackTrace();
		}		
		return output;
	}

	/**
	 * This method is also used by the global weaver editor. 
	 * @param tc
	 * @param inModel
	 * @return
	 */
	public static Map setModelsMap (ASMModel inModel, WeaverModelManager wManager, boolean isMatching, 
			boolean hasMatchConfig, ASMModel cModel, ASMModel cMetamodel ) {
		ASMAMWModel in;
		ASMAMWModel amw;
		//models
		Map models = new HashMap();
		try {
			AtlModelHandler amwModelHandler = AtlModelHandler.getDefault(TransformationConstant.AMWModelHandler);
			amw = ASMAMWModel.loadASMAMWModel("AMW", (ASMEMFModel)amwModelHandler.getMof(), wManager.getWeavingMetamodel(), null);
			models.put("AMW", amw);
			if (inModel == null) { //this means it is executed a single time, not with the configuration model 
				in = ASMAMWModel.loadASMAMWModel("IN", amw, wManager.getWeavingModel(), null);
			}
			else {
				in = ASMAMWModel.loadASMAMWModel("IN", amw, ((ASMAMWModel)inModel).getExtent(), null);
			}		
			models.put("IN", in);
			ASMModel out;
			out = ((AtlEMFModelHandler) AtlEMFModelHandler.getDefault(AtlEMFModelHandler.AMH_EMF)).newModel("OUT", "", AtlModelHandler.getDefault("AMW").getMof());
			models.put("OUT", out);
		} catch (Exception e) {
			e.printStackTrace();
		}
		return models;		
	}
	
	/**
	 * This method is also used by the global weaver editor. 
	 * @param tc
	 * @param inModel
	 * @return
	 */
	public static Map setMigratorGeneratorModelsMap (ASMModel inModel, WeaverModelManager wManager, boolean isMatching, 
			boolean hasMatchConfig, ASMModel cModel, ASMModel cMetamodel ) {
		AtlModelHandler modelHandler = AtlModelHandler.getDefault("AMW");
		ASMAMWModel in;
		ASMAMWModel amw;
		//models
		Map models = new HashMap();
		try {
			amw = ASMAMWModel.loadASMAMWModel("AMW", (ASMEMFModel)modelHandler.getMof(), wManager.getWeavingMetamodel(), null);
			models.put("AMW", amw);
			if (inModel == null) { //this means it is executed a single time, not with the configuration model 
				in = ASMAMWModel.loadASMAMWModel("COMPOSITION", amw, wManager.getWeavingModel(), null);
			}
			else {
				in = ASMAMWModel.loadASMAMWModel("COMPOSITION", amw, ((ASMAMWModel)inModel).getExtent(), null);
			}		
			models.put("COMPOSITION", in);
			ASMModel out;
			models.put("ATL", (ASMEMFModel)modelHandler.getAtl());
			out = modelHandler.newModel("OUT", "", (ASMEMFModel)modelHandler.getAtl());
			models.put("OUT", out);
		} catch (Exception e) {
			e.printStackTrace();
		}
		return models;	
	}

	/**
	 * @return the configuration object containing all the informations to execute the 'left 2 right' transformation
	 */
	public static TransformationConfig createCompositionTransformation() {
		TransformationConfig tc = new TransformationConfig () {}; 	
		tc.category = "byADL";
		tc.transformation = "transformations/Composition2Ecore.asm";
		tc.suffix = "";
		tc.description = "Compose the left and right metamodels";
		tc.bindings = new ArrayList();
		tc.hasConfigurationModel = false;
		tc.plugin="org.eclipse.weaver.extension.byADL";

		TBinding b = new TBinding () {};    				
		b.isExtraBinding =  false;	    				
		b.weavingReference = "LEFT";
		b.header = "LEFT";
		b.metamodelHeader = "MOF";
		tc.bindings.add(b);

		TBinding b2 = new TBinding () {};    				
		b2.isExtraBinding =  false;	    				
		b2.weavingReference = "RIGHT";
		b2.header = "RIGHT";
		b2.metamodelHeader = "MOF";
		tc.bindings.add(b2);
		return (tc);
	}
	
	/**
	 * @return the configuration object containing all the informations to execute the 'left 2 right' transformation
	 */
	public static TransformationConfig createSingle2ComposedGenerator() {
		TransformationConfig tc = new TransformationConfig () {}; 	
		tc.category = "byADL";
		tc.transformation = "transformations/single2ComposedMigratorGenerator.asm";
		tc.suffix = "Single2ComposedMigrator";
		tc.description = "single-2-composed migrator generator";
		tc.bindings = new ArrayList();
		tc.hasConfigurationModel = false;
		tc.plugin="org.eclipse.weaver.extension.byADL";
		
		TBinding b = new TBinding () {};    				
		b.isExtraBinding =  false;	    				
		b.weavingReference = "LEFT";
		b.header = "LEFT";
		b.metamodelHeader = "ECORE";
		tc.bindings.add(b);

		TBinding b2 = new TBinding () {};    				
		b2.isExtraBinding =  false;	    				
		b2.weavingReference = "RIGHT";
		b2.header = "RIGHT";
		b2.metamodelHeader = "ECORE";
		tc.bindings.add(b2);
		
		TBinding b3 = new TBinding () {};    				
		b3.isExtraBinding =  false;	    				
		b3.weavingReference = "IN";
		b3.header = "IN";
		b3.metamodelHeader = "ECORE";
		tc.bindings.add(b3);
		return (tc);
	}
	
	/**
	 * @return the configuration object containing all the informations to execute the 'left 2 right' transformation
	 */
	public static TransformationConfig createComposed2SingleGenerator() {
		TransformationConfig tc = new TransformationConfig () {}; 	
		tc.category = "byADL";
		tc.transformation = "transformations/composed2SingleMigratorGenerator.asm";
		tc.suffix = "Composed2SingleMigrator";
		tc.description = "composed-2-single migrator generator";
		tc.bindings = new ArrayList();
		tc.hasConfigurationModel = false;
		tc.plugin="org.eclipse.weaver.extension.byADL";
		
		TBinding b = new TBinding () {};    				
		b.isExtraBinding =  false;	    				
		b.weavingReference = "LEFT";
		b.header = "LEFT";
		b.metamodelHeader = "ECORE";
		tc.bindings.add(b);

		TBinding b2 = new TBinding () {};    				
		b2.isExtraBinding =  false;	    				
		b2.weavingReference = "RIGHT";
		b2.header = "RIGHT";
		b2.metamodelHeader = "ECORE";
		tc.bindings.add(b2);
		
		TBinding b3 = new TBinding () {};    				
		b3.isExtraBinding =  false;	    				
		b3.weavingReference = "IN";
		b3.header = "IN";
		b3.metamodelHeader = "ECORE";
		tc.bindings.add(b3);
		return (tc);
	}
	
	

	public static TransformationConfig getCompositionTransformation() {
		return TransformationsManager.compositionTransformation;
	}

	public static void setCompositionTransformation(
			TransformationConfig compositionTransformation) {
		TransformationsManager.compositionTransformation = compositionTransformation;
	}

	public static TransformationConfig getSingle2ComposedGenerator() {
		return single2ComposedGenerator;
	}

	public static void setSingle2ComposedGenerator(
			TransformationConfig single2ComposedGenerator) {
		TransformationsManager.single2ComposedGenerator = single2ComposedGenerator;
	}

	public static TransformationConfig getComposed2SingleGenerator() {
		return composed2SingleGenerator;
	}

	public static void setComposed2SingleGenerator(
			TransformationConfig composed2SingleGenerator) {
		TransformationsManager.composed2SingleGenerator = composed2SingleGenerator;
	}

	/**
	 * Adds a metamodel/profile to the internal list of the metamodels to use in the HOT transformation.
	 * 
	 * @param input the Resource to add to the internal list of the metamodels to use in the 
	 * HOT transformation.
	 */
	public static void addMetamodel(Resource input) {
		// the number of metamodels in the internal list must be always less than 2
		switch (TransformationsManager.wovenModels.size()) {
		// if there are no elements -> just add the metamodel
		case 0:	TransformationsManager.wovenModels.add(input); break;
		// if there is only one element -> just add the metamodel
		case 1: TransformationsManager.wovenModels.add(input); break;
		// if there are 2 elements -> first delete all the elements in the list, then add the metamodel
		case 2: TransformationsManager.wovenModels.clear();
		TransformationsManager.wovenModels.add(input); break;
		}
	}	

	public static void serializeMetamodel(EPackage metamodel) {
		ResourceSet metaResourceSet = new ResourceSetImpl();
		// Register XML Factory implementation to handle .ecore files
		metaResourceSet.getResourceFactoryRegistry().getExtensionToFactoryMap().put(
				"ecore", new  XMLResourceFactoryImpl());
		// Create empty resource with the given URI
		Resource metaResource = metaResourceSet.createResource(URI.createURI("./" + "testMM" + ".ecore"));
		// Add the metamodel to contents list of the resource 
		metaResource.getContents().add(metamodel);
		try {
			// Save the resource
			metaResource.save(null);
		} catch (Exception e) {
			e.printStackTrace();
		}
	}

}

